Merge branch 'for-next' of git://git.kernel.org/pub/scm/linux/kernel/git/hch/vfs...
authorLinus Torvalds <torvalds@linux-foundation.org>
Fri, 28 Oct 2011 17:49:34 +0000 (10:49 -0700)
committerLinus Torvalds <torvalds@linux-foundation.org>
Fri, 28 Oct 2011 17:49:34 +0000 (10:49 -0700)
* 'for-next' of git://git.kernel.org/pub/scm/linux/kernel/git/hch/vfs-queue: (21 commits)
  leases: fix write-open/read-lease race
  nfs: drop unnecessary locking in llseek
  ext4: replace cut'n'pasted llseek code with generic_file_llseek_size
  vfs: add generic_file_llseek_size
  vfs: do (nearly) lockless generic_file_llseek
  direct-io: merge direct_io_walker into __blockdev_direct_IO
  direct-io: inline the complete submission path
  direct-io: separate map_bh from dio
  direct-io: use a slab cache for struct dio
  direct-io: rearrange fields in dio/dio_submit to avoid holes
  direct-io: fix a wrong comment
  direct-io: separate fields only used in the submission path from struct dio
  vfs: fix spinning prevention in prune_icache_sb
  vfs: add a comment to inode_permission()
  vfs: pass all mask flags check_acl and posix_acl_permission
  vfs: add hex format for MAY_* flag values
  vfs: indicate that the permission functions take all the MAY_* flags
  compat: sync compat_stats with statfs.
  vfs: add "device" tag to /proc/self/mountstats
  cleanup: vfs: small comment fix for block_invalidatepage
  ...

Fix up trivial conflict in fs/gfs2/file.c (llseek changes)

22 files changed:
arch/mips/include/asm/compat.h
arch/parisc/include/asm/compat.h
arch/powerpc/include/asm/compat.h
arch/s390/include/asm/compat.h
arch/sparc/include/asm/compat.h
arch/x86/include/asm/compat.h
fs/btrfs/file.c
fs/buffer.c
fs/cifs/cifsfs.c
fs/compat.c
fs/direct-io.c
fs/ext4/file.c
fs/gfs2/file.c
fs/inode.c
fs/namei.c
fs/namespace.c
fs/nfs/file.c
fs/open.c
fs/posix_acl.c
fs/read_write.c
include/linux/fs.h
mm/filemap.c

index dbc5106..b77df03 100644 (file)
@@ -111,7 +111,8 @@ struct compat_statfs {
        int             f_bavail;
        compat_fsid_t   f_fsid;
        int             f_namelen;
-       int             f_spare[6];
+       int             f_flags;
+       int             f_spare[5];
 };
 
 #define COMPAT_RLIM_INFINITY   0x7fffffffUL
index efa0b60..760f331 100644 (file)
@@ -105,7 +105,8 @@ struct compat_statfs {
        __kernel_fsid_t f_fsid;
        s32             f_namelen;
        s32             f_frsize;
-       s32             f_spare[5];
+       s32             f_flags;
+       s32             f_spare[4];
 };
 
 struct compat_sigcontext {
index 91010e8..88e602f 100644 (file)
@@ -100,7 +100,8 @@ struct compat_statfs {
        compat_fsid_t   f_fsid;
        int             f_namelen;      /* SunOS ignores this field. */
        int             f_frsize;
-       int             f_spare[5];
+       int             f_flags;
+       int             f_spare[4];
 };
 
 #define COMPAT_RLIM_OLD_INFINITY       0x7fffffff
index da359ca..cdb9b78 100644 (file)
@@ -131,7 +131,8 @@ struct compat_statfs {
        compat_fsid_t   f_fsid;
        s32             f_namelen;
        s32             f_frsize;
-       s32             f_spare[6];
+       s32             f_flags;
+       s32             f_spare[5];
 };
 
 #define COMPAT_RLIM_OLD_INFINITY       0x7fffffff
index 6f57325..b8be20d 100644 (file)
@@ -134,7 +134,8 @@ struct compat_statfs {
        compat_fsid_t   f_fsid;
        int             f_namelen;      /* SunOS ignores this field. */
        int             f_frsize;
-       int             f_spare[5];
+       int             f_flags;
+       int             f_spare[4];
 };
 
 #define COMPAT_RLIM_INFINITY 0x7fffffff
index 1d9cd27..30d737e 100644 (file)
@@ -108,7 +108,8 @@ struct compat_statfs {
        compat_fsid_t   f_fsid;
        int             f_namelen;      /* SunOS ignores this field. */
        int             f_frsize;
-       int             f_spare[5];
+       int             f_flags;
+       int             f_spare[4];
 };
 
 #define COMPAT_RLIM_OLD_INFINITY       0x7fffffff
index e4e57d5..1266f6e 100644 (file)
@@ -1821,7 +1821,7 @@ static loff_t btrfs_file_llseek(struct file *file, loff_t offset, int origin)
        switch (origin) {
        case SEEK_END:
        case SEEK_CUR:
-               offset = generic_file_llseek_unlocked(file, offset, origin);
+               offset = generic_file_llseek(file, offset, origin);
                goto out;
        case SEEK_DATA:
        case SEEK_HOLE:
index 1a80b04..936d603 100644 (file)
@@ -1470,13 +1470,13 @@ static void discard_buffer(struct buffer_head * bh)
 }
 
 /**
- * block_invalidatepage - invalidate part of all of a buffer-backed page
+ * block_invalidatepage - invalidate part or all of a buffer-backed page
  *
  * @page: the page which is affected
  * @offset: the index of the truncation point
  *
  * block_invalidatepage() is called when all or part of the page has become
- * invalidatedby a truncate operation.
+ * invalidated by a truncate operation.
  *
  * block_invalidatepage() does not have to release all buffers, but it must
  * ensure that no dirty buffer is left outside @offset and that no I/O
index f219dcc..8f1fe32 100644 (file)
@@ -730,7 +730,7 @@ static loff_t cifs_llseek(struct file *file, loff_t offset, int origin)
                if (rc < 0)
                        return (loff_t)rc;
        }
-       return generic_file_llseek_unlocked(file, offset, origin);
+       return generic_file_llseek(file, offset, origin);
 }
 
 static int cifs_setlease(struct file *file, long arg, struct file_lock **lease)
index 05e3f3d..302e761 100644 (file)
@@ -246,11 +246,8 @@ static int put_compat_statfs(struct compat_statfs __user *ubuf, struct kstatfs *
            __put_user(kbuf->f_fsid.val[0], &ubuf->f_fsid.val[0]) ||
            __put_user(kbuf->f_fsid.val[1], &ubuf->f_fsid.val[1]) ||
            __put_user(kbuf->f_frsize, &ubuf->f_frsize) ||
-           __put_user(0, &ubuf->f_spare[0]) || 
-           __put_user(0, &ubuf->f_spare[1]) || 
-           __put_user(0, &ubuf->f_spare[2]) || 
-           __put_user(0, &ubuf->f_spare[3]) || 
-           __put_user(0, &ubuf->f_spare[4]))
+           __put_user(kbuf->f_flags, &ubuf->f_flags) ||
+           __clear_user(ubuf->f_spare, sizeof(ubuf->f_spare)))
                return -EFAULT;
        return 0;
 }
index 44a360c..d740ab6 100644 (file)
@@ -39,7 +39,7 @@
 
 /*
  * How many user pages to map in one call to get_user_pages().  This determines
- * the size of a structure on the stack.
+ * the size of a structure in the slab cache
  */
 #define DIO_PAGES      64
 
  * blocksize.
  */
 
-struct dio {
-       /* BIO submission state */
+/* dio_state only used in the submission path */
+
+struct dio_submit {
        struct bio *bio;                /* bio under assembly */
-       struct inode *inode;
-       int rw;
-       loff_t i_size;                  /* i_size when submitted */
-       int flags;                      /* doesn't change */
        unsigned blkbits;               /* doesn't change */
        unsigned blkfactor;             /* When we're using an alignment which
                                           is finer than the filesystem's soft
@@ -76,18 +73,17 @@ struct dio {
        sector_t block_in_file;         /* Current offset into the underlying
                                           file in dio_block units. */
        unsigned blocks_available;      /* At block_in_file.  changes */
+       int reap_counter;               /* rate limit reaping */
        sector_t final_block_in_request;/* doesn't change */
        unsigned first_block_in_page;   /* doesn't change, Used only once */
        int boundary;                   /* prev block is at a boundary */
-       int reap_counter;               /* rate limit reaping */
        get_block_t *get_block;         /* block mapping function */
-       dio_iodone_t *end_io;           /* IO completion function */
        dio_submit_t *submit_io;        /* IO submition function */
+
        loff_t logical_offset_in_bio;   /* current first logical block in bio */
        sector_t final_block_in_bio;    /* current final block in bio + 1 */
        sector_t next_block_for_io;     /* next block to be put under IO,
                                           in dio_blocks units */
-       struct buffer_head map_bh;      /* last get_block() result */
 
        /*
         * Deferred addition of a page to the dio.  These variables are
@@ -100,18 +96,6 @@ struct dio {
        sector_t cur_page_block;        /* Where it starts */
        loff_t cur_page_fs_offset;      /* Offset in file */
 
-       /* BIO completion state */
-       spinlock_t bio_lock;            /* protects BIO fields below */
-       unsigned long refcount;         /* direct_io_worker() and bios */
-       struct bio *bio_list;           /* singly linked via bi_private */
-       struct task_struct *waiter;     /* waiting task (NULL if none) */
-
-       /* AIO related stuff */
-       struct kiocb *iocb;             /* kiocb */
-       int is_async;                   /* is IO async ? */
-       int io_error;                   /* IO error in completion path */
-       ssize_t result;                 /* IO result */
-
        /*
         * Page fetching state. These variables belong to dio_refill_pages().
         */
@@ -125,7 +109,30 @@ struct dio {
         */
        unsigned head;                  /* next page to process */
        unsigned tail;                  /* last valid page + 1 */
+};
+
+/* dio_state communicated between submission path and end_io */
+struct dio {
+       int flags;                      /* doesn't change */
+       int rw;
+       struct inode *inode;
+       loff_t i_size;                  /* i_size when submitted */
+       dio_iodone_t *end_io;           /* IO completion function */
+
+       void *private;                  /* copy from map_bh.b_private */
+
+       /* BIO completion state */
+       spinlock_t bio_lock;            /* protects BIO fields below */
        int page_errors;                /* errno from get_user_pages() */
+       int is_async;                   /* is IO async ? */
+       int io_error;                   /* IO error in completion path */
+       unsigned long refcount;         /* direct_io_worker() and bios */
+       struct bio *bio_list;           /* singly linked via bi_private */
+       struct task_struct *waiter;     /* waiting task (NULL if none) */
+
+       /* AIO related stuff */
+       struct kiocb *iocb;             /* kiocb */
+       ssize_t result;                 /* IO result */
 
        /*
         * pages[] (and any fields placed after it) are not zeroed out at
@@ -133,7 +140,9 @@ struct dio {
         * wish that they not be zeroed.
         */
        struct page *pages[DIO_PAGES];  /* page buffer */
-};
+} ____cacheline_aligned_in_smp;
+
+static struct kmem_cache *dio_cache __read_mostly;
 
 static void __inode_dio_wait(struct inode *inode)
 {
@@ -182,27 +191,27 @@ EXPORT_SYMBOL_GPL(inode_dio_done);
 /*
  * How many pages are in the queue?
  */
-static inline unsigned dio_pages_present(struct dio *dio)
+static inline unsigned dio_pages_present(struct dio_submit *sdio)
 {
-       return dio->tail - dio->head;
+       return sdio->tail - sdio->head;
 }
 
 /*
  * Go grab and pin some userspace pages.   Typically we'll get 64 at a time.
  */
-static int dio_refill_pages(struct dio *dio)
+static inline int dio_refill_pages(struct dio *dio, struct dio_submit *sdio)
 {
        int ret;
        int nr_pages;
 
-       nr_pages = min(dio->total_pages - dio->curr_page, DIO_PAGES);
+       nr_pages = min(sdio->total_pages - sdio->curr_page, DIO_PAGES);
        ret = get_user_pages_fast(
-               dio->curr_user_address,         /* Where from? */
+               sdio->curr_user_address,                /* Where from? */
                nr_pages,                       /* How many pages? */
                dio->rw == READ,                /* Write to memory? */
                &dio->pages[0]);                /* Put results here */
 
-       if (ret < 0 && dio->blocks_available && (dio->rw & WRITE)) {
+       if (ret < 0 && sdio->blocks_available && (dio->rw & WRITE)) {
                struct page *page = ZERO_PAGE(0);
                /*
                 * A memory fault, but the filesystem has some outstanding
@@ -213,17 +222,17 @@ static int dio_refill_pages(struct dio *dio)
                        dio->page_errors = ret;
                page_cache_get(page);
                dio->pages[0] = page;
-               dio->head = 0;
-               dio->tail = 1;
+               sdio->head = 0;
+               sdio->tail = 1;
                ret = 0;
                goto out;
        }
 
        if (ret >= 0) {
-               dio->curr_user_address += ret * PAGE_SIZE;
-               dio->curr_page += ret;
-               dio->head = 0;
-               dio->tail = ret;
+               sdio->curr_user_address += ret * PAGE_SIZE;
+               sdio->curr_page += ret;
+               sdio->head = 0;
+               sdio->tail = ret;
                ret = 0;
        }
 out:
@@ -236,17 +245,18 @@ out:
  * decent number of pages, less frequently.  To provide nicer use of the
  * L1 cache.
  */
-static struct page *dio_get_page(struct dio *dio)
+static inline struct page *dio_get_page(struct dio *dio,
+               struct dio_submit *sdio)
 {
-       if (dio_pages_present(dio) == 0) {
+       if (dio_pages_present(sdio) == 0) {
                int ret;
 
-               ret = dio_refill_pages(dio);
+               ret = dio_refill_pages(dio, sdio);
                if (ret)
                        return ERR_PTR(ret);
-               BUG_ON(dio_pages_present(dio) == 0);
+               BUG_ON(dio_pages_present(sdio) == 0);
        }
-       return dio->pages[dio->head++];
+       return dio->pages[sdio->head++];
 }
 
 /**
@@ -292,7 +302,7 @@ static ssize_t dio_complete(struct dio *dio, loff_t offset, ssize_t ret, bool is
 
        if (dio->end_io && dio->result) {
                dio->end_io(dio->iocb, offset, transferred,
-                           dio->map_bh.b_private, ret, is_async);
+                           dio->private, ret, is_async);
        } else {
                if (is_async)
                        aio_complete(dio->iocb, ret, 0);
@@ -323,7 +333,7 @@ static void dio_bio_end_aio(struct bio *bio, int error)
 
        if (remaining == 0) {
                dio_complete(dio, dio->iocb->ki_pos, 0, true);
-               kfree(dio);
+               kmem_cache_free(dio_cache, dio);
        }
 }
 
@@ -367,9 +377,10 @@ void dio_end_io(struct bio *bio, int error)
 }
 EXPORT_SYMBOL_GPL(dio_end_io);
 
-static void
-dio_bio_alloc(struct dio *dio, struct block_device *bdev,
-               sector_t first_sector, int nr_vecs)
+static inline void
+dio_bio_alloc(struct dio *dio, struct dio_submit *sdio,
+             struct block_device *bdev,
+             sector_t first_sector, int nr_vecs)
 {
        struct bio *bio;
 
@@ -386,8 +397,8 @@ dio_bio_alloc(struct dio *dio, struct block_device *bdev,
        else
                bio->bi_end_io = dio_bio_end_io;
 
-       dio->bio = bio;
-       dio->logical_offset_in_bio = dio->cur_page_fs_offset;
+       sdio->bio = bio;
+       sdio->logical_offset_in_bio = sdio->cur_page_fs_offset;
 }
 
 /*
@@ -397,9 +408,9 @@ dio_bio_alloc(struct dio *dio, struct block_device *bdev,
  *
  * bios hold a dio reference between submit_bio and ->end_io.
  */
-static void dio_bio_submit(struct dio *dio)
+static inline void dio_bio_submit(struct dio *dio, struct dio_submit *sdio)
 {
-       struct bio *bio = dio->bio;
+       struct bio *bio = sdio->bio;
        unsigned long flags;
 
        bio->bi_private = dio;
@@ -411,24 +422,24 @@ static void dio_bio_submit(struct dio *dio)
        if (dio->is_async && dio->rw == READ)
                bio_set_pages_dirty(bio);
 
-       if (dio->submit_io)
-               dio->submit_io(dio->rw, bio, dio->inode,
-                              dio->logical_offset_in_bio);
+       if (sdio->submit_io)
+               sdio->submit_io(dio->rw, bio, dio->inode,
+                              sdio->logical_offset_in_bio);
        else
                submit_bio(dio->rw, bio);
 
-       dio->bio = NULL;
-       dio->boundary = 0;
-       dio->logical_offset_in_bio = 0;
+       sdio->bio = NULL;
+       sdio->boundary = 0;
+       sdio->logical_offset_in_bio = 0;
 }
 
 /*
  * Release any resources in case of a failure
  */
-static void dio_cleanup(struct dio *dio)
+static inline void dio_cleanup(struct dio *dio, struct dio_submit *sdio)
 {
-       while (dio_pages_present(dio))
-               page_cache_release(dio_get_page(dio));
+       while (dio_pages_present(sdio))
+               page_cache_release(dio_get_page(dio, sdio));
 }
 
 /*
@@ -518,11 +529,11 @@ static void dio_await_completion(struct dio *dio)
  *
  * This also helps to limit the peak amount of pinned userspace memory.
  */
-static int dio_bio_reap(struct dio *dio)
+static inline int dio_bio_reap(struct dio *dio, struct dio_submit *sdio)
 {
        int ret = 0;
 
-       if (dio->reap_counter++ >= 64) {
+       if (sdio->reap_counter++ >= 64) {
                while (dio->bio_list) {
                        unsigned long flags;
                        struct bio *bio;
@@ -536,14 +547,14 @@ static int dio_bio_reap(struct dio *dio)
                        if (ret == 0)
                                ret = ret2;
                }
-               dio->reap_counter = 0;
+               sdio->reap_counter = 0;
        }
        return ret;
 }
 
 /*
  * Call into the fs to map some more disk blocks.  We record the current number
- * of available blocks at dio->blocks_available.  These are in units of the
+ * of available blocks at sdio->blocks_available.  These are in units of the
  * fs blocksize, (1 << inode->i_blkbits).
  *
  * The fs is allowed to map lots of blocks at once.  If it wants to do that,
@@ -564,10 +575,10 @@ static int dio_bio_reap(struct dio *dio)
  * buffer_mapped().  However the direct-io code will only process holes one
  * block at a time - it will repeatedly call get_block() as it walks the hole.
  */
-static int get_more_blocks(struct dio *dio)
+static int get_more_blocks(struct dio *dio, struct dio_submit *sdio,
+                          struct buffer_head *map_bh)
 {
        int ret;
-       struct buffer_head *map_bh = &dio->map_bh;
        sector_t fs_startblk;   /* Into file, in filesystem-sized blocks */
        unsigned long fs_count; /* Number of filesystem-sized blocks */
        unsigned long dio_count;/* Number of dio_block-sized blocks */
@@ -580,11 +591,11 @@ static int get_more_blocks(struct dio *dio)
         */
        ret = dio->page_errors;
        if (ret == 0) {
-               BUG_ON(dio->block_in_file >= dio->final_block_in_request);
-               fs_startblk = dio->block_in_file >> dio->blkfactor;
-               dio_count = dio->final_block_in_request - dio->block_in_file;
-               fs_count = dio_count >> dio->blkfactor;
-               blkmask = (1 << dio->blkfactor) - 1;
+               BUG_ON(sdio->block_in_file >= sdio->final_block_in_request);
+               fs_startblk = sdio->block_in_file >> sdio->blkfactor;
+               dio_count = sdio->final_block_in_request - sdio->block_in_file;
+               fs_count = dio_count >> sdio->blkfactor;
+               blkmask = (1 << sdio->blkfactor) - 1;
                if (dio_count & blkmask)        
                        fs_count++;
 
@@ -604,13 +615,16 @@ static int get_more_blocks(struct dio *dio)
                 */
                create = dio->rw & WRITE;
                if (dio->flags & DIO_SKIP_HOLES) {
-                       if (dio->block_in_file < (i_size_read(dio->inode) >>
-                                                       dio->blkbits))
+                       if (sdio->block_in_file < (i_size_read(dio->inode) >>
+                                                       sdio->blkbits))
                                create = 0;
                }
 
-               ret = (*dio->get_block)(dio->inode, fs_startblk,
+               ret = (*sdio->get_block)(dio->inode, fs_startblk,
                                                map_bh, create);
+
+               /* Store for completion */
+               dio->private = map_bh->b_private;
        }
        return ret;
 }
@@ -618,20 +632,21 @@ static int get_more_blocks(struct dio *dio)
 /*
  * There is no bio.  Make one now.
  */
-static int dio_new_bio(struct dio *dio, sector_t start_sector)
+static inline int dio_new_bio(struct dio *dio, struct dio_submit *sdio,
+               sector_t start_sector, struct buffer_head *map_bh)
 {
        sector_t sector;
        int ret, nr_pages;
 
-       ret = dio_bio_reap(dio);
+       ret = dio_bio_reap(dio, sdio);
        if (ret)
                goto out;
-       sector = start_sector << (dio->blkbits - 9);
-       nr_pages = min(dio->pages_in_io, bio_get_nr_vecs(dio->map_bh.b_bdev));
+       sector = start_sector << (sdio->blkbits - 9);
+       nr_pages = min(sdio->pages_in_io, bio_get_nr_vecs(map_bh->b_bdev));
        nr_pages = min(nr_pages, BIO_MAX_PAGES);
        BUG_ON(nr_pages <= 0);
-       dio_bio_alloc(dio, dio->map_bh.b_bdev, sector, nr_pages);
-       dio->boundary = 0;
+       dio_bio_alloc(dio, sdio, map_bh->b_bdev, sector, nr_pages);
+       sdio->boundary = 0;
 out:
        return ret;
 }
@@ -643,21 +658,21 @@ out:
  *
  * Return zero on success.  Non-zero means the caller needs to start a new BIO.
  */
-static int dio_bio_add_page(struct dio *dio)
+static inline int dio_bio_add_page(struct dio_submit *sdio)
 {
        int ret;
 
-       ret = bio_add_page(dio->bio, dio->cur_page,
-                       dio->cur_page_len, dio->cur_page_offset);
-       if (ret == dio->cur_page_len) {
+       ret = bio_add_page(sdio->bio, sdio->cur_page,
+                       sdio->cur_page_len, sdio->cur_page_offset);
+       if (ret == sdio->cur_page_len) {
                /*
                 * Decrement count only, if we are done with this page
                 */
-               if ((dio->cur_page_len + dio->cur_page_offset) == PAGE_SIZE)
-                       dio->pages_in_io--;
-               page_cache_get(dio->cur_page);
-               dio->final_block_in_bio = dio->cur_page_block +
-                       (dio->cur_page_len >> dio->blkbits);
+               if ((sdio->cur_page_len + sdio->cur_page_offset) == PAGE_SIZE)
+                       sdio->pages_in_io--;
+               page_cache_get(sdio->cur_page);
+               sdio->final_block_in_bio = sdio->cur_page_block +
+                       (sdio->cur_page_len >> sdio->blkbits);
                ret = 0;
        } else {
                ret = 1;
@@ -675,14 +690,15 @@ static int dio_bio_add_page(struct dio *dio)
  * The caller of this function is responsible for removing cur_page from the
  * dio, and for dropping the refcount which came from that presence.
  */
-static int dio_send_cur_page(struct dio *dio)
+static inline int dio_send_cur_page(struct dio *dio, struct dio_submit *sdio,
+               struct buffer_head *map_bh)
 {
        int ret = 0;
 
-       if (dio->bio) {
-               loff_t cur_offset = dio->cur_page_fs_offset;
-               loff_t bio_next_offset = dio->logical_offset_in_bio +
-                       dio->bio->bi_size;
+       if (sdio->bio) {
+               loff_t cur_offset = sdio->cur_page_fs_offset;
+               loff_t bio_next_offset = sdio->logical_offset_in_bio +
+                       sdio->bio->bi_size;
 
                /*
                 * See whether this new request is contiguous with the old.
@@ -698,28 +714,28 @@ static int dio_send_cur_page(struct dio *dio)
                 * be the next logical offset in the bio, submit the bio we
                 * have.
                 */
-               if (dio->final_block_in_bio != dio->cur_page_block ||
+               if (sdio->final_block_in_bio != sdio->cur_page_block ||
                    cur_offset != bio_next_offset)
-                       dio_bio_submit(dio);
+                       dio_bio_submit(dio, sdio);
                /*
                 * Submit now if the underlying fs is about to perform a
                 * metadata read
                 */
-               else if (dio->boundary)
-                       dio_bio_submit(dio);
+               else if (sdio->boundary)
+                       dio_bio_submit(dio, sdio);
        }
 
-       if (dio->bio == NULL) {
-               ret = dio_new_bio(dio, dio->cur_page_block);
+       if (sdio->bio == NULL) {
+               ret = dio_new_bio(dio, sdio, sdio->cur_page_block, map_bh);
                if (ret)
                        goto out;
        }
 
-       if (dio_bio_add_page(dio) != 0) {
-               dio_bio_submit(dio);
-               ret = dio_new_bio(dio, dio->cur_page_block);
+       if (dio_bio_add_page(sdio) != 0) {
+               dio_bio_submit(dio, sdio);
+               ret = dio_new_bio(dio, sdio, sdio->cur_page_block, map_bh);
                if (ret == 0) {
-                       ret = dio_bio_add_page(dio);
+                       ret = dio_bio_add_page(sdio);
                        BUG_ON(ret != 0);
                }
        }
@@ -744,9 +760,10 @@ out:
  * If that doesn't work out then we put the old page into the bio and add this
  * page to the dio instead.
  */
-static int
-submit_page_section(struct dio *dio, struct page *page,
-               unsigned offset, unsigned len, sector_t blocknr)
+static inline int
+submit_page_section(struct dio *dio, struct dio_submit *sdio, struct page *page,
+                   unsigned offset, unsigned len, sector_t blocknr,
+                   struct buffer_head *map_bh)
 {
        int ret = 0;
 
@@ -760,20 +777,20 @@ submit_page_section(struct dio *dio, struct page *page,
        /*
         * Can we just grow the current page's presence in the dio?
         */
-       if (    (dio->cur_page == page) &&
-               (dio->cur_page_offset + dio->cur_page_len == offset) &&
-               (dio->cur_page_block +
-                       (dio->cur_page_len >> dio->blkbits) == blocknr)) {
-               dio->cur_page_len += len;
+       if (sdio->cur_page == page &&
+           sdio->cur_page_offset + sdio->cur_page_len == offset &&
+           sdio->cur_page_block +
+           (sdio->cur_page_len >> sdio->blkbits) == blocknr) {
+               sdio->cur_page_len += len;
 
                /*
-                * If dio->boundary then we want to schedule the IO now to
+                * If sdio->boundary then we want to schedule the IO now to
                 * avoid metadata seeks.
                 */
-               if (dio->boundary) {
-                       ret = dio_send_cur_page(dio);
-                       page_cache_release(dio->cur_page);
-                       dio->cur_page = NULL;
+               if (sdio->boundary) {
+                       ret = dio_send_cur_page(dio, sdio, map_bh);
+                       page_cache_release(sdio->cur_page);
+                       sdio->cur_page = NULL;
                }
                goto out;
        }
@@ -781,20 +798,20 @@ submit_page_section(struct dio *dio, struct page *page,
        /*
         * If there's a deferred page already there then send it.
         */
-       if (dio->cur_page) {
-               ret = dio_send_cur_page(dio);
-               page_cache_release(dio->cur_page);
-               dio->cur_page = NULL;
+       if (sdio->cur_page) {
+               ret = dio_send_cur_page(dio, sdio, map_bh);
+               page_cache_release(sdio->cur_page);
+               sdio->cur_page = NULL;
                if (ret)
                        goto out;
        }
 
        page_cache_get(page);           /* It is in dio */
-       dio->cur_page = page;
-       dio->cur_page_offset = offset;
-       dio->cur_page_len = len;
-       dio->cur_page_block = blocknr;
-       dio->cur_page_fs_offset = dio->block_in_file << dio->blkbits;
+       sdio->cur_page = page;
+       sdio->cur_page_offset = offset;
+       sdio->cur_page_len = len;
+       sdio->cur_page_block = blocknr;
+       sdio->cur_page_fs_offset = sdio->block_in_file << sdio->blkbits;
 out:
        return ret;
 }
@@ -804,16 +821,16 @@ out:
  * file blocks.  Only called for S_ISREG files - blockdevs do not set
  * buffer_new
  */
-static void clean_blockdev_aliases(struct dio *dio)
+static void clean_blockdev_aliases(struct dio *dio, struct buffer_head *map_bh)
 {
        unsigned i;
        unsigned nblocks;
 
-       nblocks = dio->map_bh.b_size >> dio->inode->i_blkbits;
+       nblocks = map_bh->b_size >> dio->inode->i_blkbits;
 
        for (i = 0; i < nblocks; i++) {
-               unmap_underlying_metadata(dio->map_bh.b_bdev,
-                                       dio->map_bh.b_blocknr + i);
+               unmap_underlying_metadata(map_bh->b_bdev,
+                                         map_bh->b_blocknr + i);
        }
 }
 
@@ -826,19 +843,20 @@ static void clean_blockdev_aliases(struct dio *dio)
  * `end' is zero if we're doing the start of the IO, 1 at the end of the
  * IO.
  */
-static void dio_zero_block(struct dio *dio, int end)
+static inline void dio_zero_block(struct dio *dio, struct dio_submit *sdio,
+               int end, struct buffer_head *map_bh)
 {
        unsigned dio_blocks_per_fs_block;
        unsigned this_chunk_blocks;     /* In dio_blocks */
        unsigned this_chunk_bytes;
        struct page *page;
 
-       dio->start_zero_done = 1;
-       if (!dio->blkfactor || !buffer_new(&dio->map_bh))
+       sdio->start_zero_done = 1;
+       if (!sdio->blkfactor || !buffer_new(map_bh))
                return;
 
-       dio_blocks_per_fs_block = 1 << dio->blkfactor;
-       this_chunk_blocks = dio->block_in_file & (dio_blocks_per_fs_block - 1);
+       dio_blocks_per_fs_block = 1 << sdio->blkfactor;
+       this_chunk_blocks = sdio->block_in_file & (dio_blocks_per_fs_block - 1);
 
        if (!this_chunk_blocks)
                return;
@@ -850,14 +868,14 @@ static void dio_zero_block(struct dio *dio, int end)
        if (end) 
                this_chunk_blocks = dio_blocks_per_fs_block - this_chunk_blocks;
 
-       this_chunk_bytes = this_chunk_blocks << dio->blkbits;
+       this_chunk_bytes = this_chunk_blocks << sdio->blkbits;
 
        page = ZERO_PAGE(0);
-       if (submit_page_section(dio, page, 0, this_chunk_bytes, 
-                               dio->next_block_for_io))
+       if (submit_page_section(dio, sdio, page, 0, this_chunk_bytes,
+                               sdio->next_block_for_io, map_bh))
                return;
 
-       dio->next_block_for_io += this_chunk_blocks;
+       sdio->next_block_for_io += this_chunk_blocks;
 }
 
 /*
@@ -876,20 +894,20 @@ static void dio_zero_block(struct dio *dio, int end)
  * it should set b_size to PAGE_SIZE or more inside get_block().  This gives
  * fine alignment but still allows this function to work in PAGE_SIZE units.
  */
-static int do_direct_IO(struct dio *dio)
+static int do_direct_IO(struct dio *dio, struct dio_submit *sdio,
+                       struct buffer_head *map_bh)
 {
-       const unsigned blkbits = dio->blkbits;
+       const unsigned blkbits = sdio->blkbits;
        const unsigned blocks_per_page = PAGE_SIZE >> blkbits;
        struct page *page;
        unsigned block_in_page;
-       struct buffer_head *map_bh = &dio->map_bh;
        int ret = 0;
 
        /* The I/O can start at any block offset within the first page */
-       block_in_page = dio->first_block_in_page;
+       block_in_page = sdio->first_block_in_page;
 
-       while (dio->block_in_file < dio->final_block_in_request) {
-               page = dio_get_page(dio);
+       while (sdio->block_in_file < sdio->final_block_in_request) {
+               page = dio_get_page(dio, sdio);
                if (IS_ERR(page)) {
                        ret = PTR_ERR(page);
                        goto out;
@@ -901,14 +919,14 @@ static int do_direct_IO(struct dio *dio)
                        unsigned this_chunk_blocks;     /* # of blocks */
                        unsigned u;
 
-                       if (dio->blocks_available == 0) {
+                       if (sdio->blocks_available == 0) {
                                /*
                                 * Need to go and map some more disk
                                 */
                                unsigned long blkmask;
                                unsigned long dio_remainder;
 
-                               ret = get_more_blocks(dio);
+                               ret = get_more_blocks(dio, sdio, map_bh);
                                if (ret) {
                                        page_cache_release(page);
                                        goto out;
@@ -916,18 +934,18 @@ static int do_direct_IO(struct dio *dio)
                                if (!buffer_mapped(map_bh))
                                        goto do_holes;
 
-                               dio->blocks_available =
-                                               map_bh->b_size >> dio->blkbits;
-                               dio->next_block_for_io =
-                                       map_bh->b_blocknr << dio->blkfactor;
+                               sdio->blocks_available =
+                                               map_bh->b_size >> sdio->blkbits;
+                               sdio->next_block_for_io =
+                                       map_bh->b_blocknr << sdio->blkfactor;
                                if (buffer_new(map_bh))
-                                       clean_blockdev_aliases(dio);
+                                       clean_blockdev_aliases(dio, map_bh);
 
-                               if (!dio->blkfactor)
+                               if (!sdio->blkfactor)
                                        goto do_holes;
 
-                               blkmask = (1 << dio->blkfactor) - 1;
-                               dio_remainder = (dio->block_in_file & blkmask);
+                               blkmask = (1 << sdio->blkfactor) - 1;
+                               dio_remainder = (sdio->block_in_file & blkmask);
 
                                /*
                                 * If we are at the start of IO and that IO
@@ -941,8 +959,8 @@ static int do_direct_IO(struct dio *dio)
                                 * on-disk
                                 */
                                if (!buffer_new(map_bh))
-                                       dio->next_block_for_io += dio_remainder;
-                               dio->blocks_available -= dio_remainder;
+                                       sdio->next_block_for_io += dio_remainder;
+                               sdio->blocks_available -= dio_remainder;
                        }
 do_holes:
                        /* Handle holes */
@@ -961,7 +979,7 @@ do_holes:
                                 */
                                i_size_aligned = ALIGN(i_size_read(dio->inode),
                                                        1 << blkbits);
-                               if (dio->block_in_file >=
+                               if (sdio->block_in_file >=
                                                i_size_aligned >> blkbits) {
                                        /* We hit eof */
                                        page_cache_release(page);
@@ -969,7 +987,7 @@ do_holes:
                                }
                                zero_user(page, block_in_page << blkbits,
                                                1 << blkbits);
-                               dio->block_in_file++;
+                               sdio->block_in_file++;
                                block_in_page++;
                                goto next_block;
                        }
@@ -979,38 +997,41 @@ do_holes:
                         * is finer than the underlying fs, go check to see if
                         * we must zero out the start of this block.
                         */
-                       if (unlikely(dio->blkfactor && !dio->start_zero_done))
-                               dio_zero_block(dio, 0);
+                       if (unlikely(sdio->blkfactor && !sdio->start_zero_done))
+                               dio_zero_block(dio, sdio, 0, map_bh);
 
                        /*
                         * Work out, in this_chunk_blocks, how much disk we
                         * can add to this page
                         */
-                       this_chunk_blocks = dio->blocks_available;
+                       this_chunk_blocks = sdio->blocks_available;
                        u = (PAGE_SIZE - offset_in_page) >> blkbits;
                        if (this_chunk_blocks > u)
                                this_chunk_blocks = u;
-                       u = dio->final_block_in_request - dio->block_in_file;
+                       u = sdio->final_block_in_request - sdio->block_in_file;
                        if (this_chunk_blocks > u)
                                this_chunk_blocks = u;
                        this_chunk_bytes = this_chunk_blocks << blkbits;
                        BUG_ON(this_chunk_bytes == 0);
 
-                       dio->boundary = buffer_boundary(map_bh);
-                       ret = submit_page_section(dio, page, offset_in_page,
-                               this_chunk_bytes, dio->next_block_for_io);
+                       sdio->boundary = buffer_boundary(map_bh);
+                       ret = submit_page_section(dio, sdio, page,
+                                                 offset_in_page,
+                                                 this_chunk_bytes,
+                                                 sdio->next_block_for_io,
+                                                 map_bh);
                        if (ret) {
                                page_cache_release(page);
                                goto out;
                        }
-                       dio->next_block_for_io += this_chunk_blocks;
+                       sdio->next_block_for_io += this_chunk_blocks;
 
-                       dio->block_in_file += this_chunk_blocks;
+                       sdio->block_in_file += this_chunk_blocks;
                        block_in_page += this_chunk_blocks;
-                       dio->blocks_available -= this_chunk_blocks;
+                       sdio->blocks_available -= this_chunk_blocks;
 next_block:
-                       BUG_ON(dio->block_in_file > dio->final_block_in_request);
-                       if (dio->block_in_file == dio->final_block_in_request)
+                       BUG_ON(sdio->block_in_file > sdio->final_block_in_request);
+                       if (sdio->block_in_file == sdio->final_block_in_request)
                                break;
                }
 
@@ -1022,135 +1043,10 @@ out:
        return ret;
 }
 
-static ssize_t
-direct_io_worker(int rw, struct kiocb *iocb, struct inode *inode, 
-       const struct iovec *iov, loff_t offset, unsigned long nr_segs, 
-       unsigned blkbits, get_block_t get_block, dio_iodone_t end_io,
-       dio_submit_t submit_io, struct dio *dio)
+static inline int drop_refcount(struct dio *dio)
 {
-       unsigned long user_addr; 
+       int ret2;
        unsigned long flags;
-       int seg;
-       ssize_t ret = 0;
-       ssize_t ret2;
-       size_t bytes;
-
-       dio->inode = inode;
-       dio->rw = rw;
-       dio->blkbits = blkbits;
-       dio->blkfactor = inode->i_blkbits - blkbits;
-       dio->block_in_file = offset >> blkbits;
-
-       dio->get_block = get_block;
-       dio->end_io = end_io;
-       dio->submit_io = submit_io;
-       dio->final_block_in_bio = -1;
-       dio->next_block_for_io = -1;
-
-       dio->iocb = iocb;
-       dio->i_size = i_size_read(inode);
-
-       spin_lock_init(&dio->bio_lock);
-       dio->refcount = 1;
-
-       /*
-        * In case of non-aligned buffers, we may need 2 more
-        * pages since we need to zero out first and last block.
-        */
-       if (unlikely(dio->blkfactor))
-               dio->pages_in_io = 2;
-
-       for (seg = 0; seg < nr_segs; seg++) {
-               user_addr = (unsigned long)iov[seg].iov_base;
-               dio->pages_in_io +=
-                       ((user_addr+iov[seg].iov_len +PAGE_SIZE-1)/PAGE_SIZE
-                               - user_addr/PAGE_SIZE);
-       }
-
-       for (seg = 0; seg < nr_segs; seg++) {
-               user_addr = (unsigned long)iov[seg].iov_base;
-               dio->size += bytes = iov[seg].iov_len;
-
-               /* Index into the first page of the first block */
-               dio->first_block_in_page = (user_addr & ~PAGE_MASK) >> blkbits;
-               dio->final_block_in_request = dio->block_in_file +
-                                               (bytes >> blkbits);
-               /* Page fetching state */
-               dio->head = 0;
-               dio->tail = 0;
-               dio->curr_page = 0;
-
-               dio->total_pages = 0;
-               if (user_addr & (PAGE_SIZE-1)) {
-                       dio->total_pages++;
-                       bytes -= PAGE_SIZE - (user_addr & (PAGE_SIZE - 1));
-               }
-               dio->total_pages += (bytes + PAGE_SIZE - 1) / PAGE_SIZE;
-               dio->curr_user_address = user_addr;
-       
-               ret = do_direct_IO(dio);
-
-               dio->result += iov[seg].iov_len -
-                       ((dio->final_block_in_request - dio->block_in_file) <<
-                                       blkbits);
-
-               if (ret) {
-                       dio_cleanup(dio);
-                       break;
-               }
-       } /* end iovec loop */
-
-       if (ret == -ENOTBLK) {
-               /*
-                * The remaining part of the request will be
-                * be handled by buffered I/O when we return
-                */
-               ret = 0;
-       }
-       /*
-        * There may be some unwritten disk at the end of a part-written
-        * fs-block-sized block.  Go zero that now.
-        */
-       dio_zero_block(dio, 1);
-
-       if (dio->cur_page) {
-               ret2 = dio_send_cur_page(dio);
-               if (ret == 0)
-                       ret = ret2;
-               page_cache_release(dio->cur_page);
-               dio->cur_page = NULL;
-       }
-       if (dio->bio)
-               dio_bio_submit(dio);
-
-       /*
-        * It is possible that, we return short IO due to end of file.
-        * In that case, we need to release all the pages we got hold on.
-        */
-       dio_cleanup(dio);
-
-       /*
-        * All block lookups have been performed. For READ requests
-        * we can let i_mutex go now that its achieved its purpose
-        * of protecting us from looking up uninitialized blocks.
-        */
-       if (rw == READ && (dio->flags & DIO_LOCKING))
-               mutex_unlock(&dio->inode->i_mutex);
-
-       /*
-        * The only time we want to leave bios in flight is when a successful
-        * partial aio read or full aio write have been setup.  In that case
-        * bio completion will call aio_complete.  The only time it's safe to
-        * call aio_complete is when we return -EIOCBQUEUED, so we key on that.
-        * This had *better* be the only place that raises -EIOCBQUEUED.
-        */
-       BUG_ON(ret == -EIOCBQUEUED);
-       if (dio->is_async && ret == 0 && dio->result &&
-           ((rw & READ) || (dio->result == dio->size)))
-               ret = -EIOCBQUEUED;
-
-       if (ret != -EIOCBQUEUED)
-               dio_await_completion(dio);
 
        /*
         * Sync will always be dropping the final ref and completing the
@@ -1166,14 +1062,7 @@ direct_io_worker(int rw, struct kiocb *iocb, struct inode *inode,
        spin_lock_irqsave(&dio->bio_lock, flags);
        ret2 = --dio->refcount;
        spin_unlock_irqrestore(&dio->bio_lock, flags);
-
-       if (ret2 == 0) {
-               ret = dio_complete(dio, offset, ret, false);
-               kfree(dio);
-       } else
-               BUG_ON(ret != -EIOCBQUEUED);
-
-       return ret;
+       return ret2;
 }
 
 /*
@@ -1195,6 +1084,11 @@ direct_io_worker(int rw, struct kiocb *iocb, struct inode *inode,
  * expected that filesystem provide exclusion between new direct I/O
  * and truncates.  For DIO_LOCKING filesystems this is done by i_mutex,
  * but other filesystems need to take care of this on their own.
+ *
+ * NOTE: if you pass "sdio" to anything by pointer make sure that function
+ * is always inlined. Otherwise gcc is unable to split the structure into
+ * individual fields and will generate much worse code. This is important
+ * for the whole file.
  */
 ssize_t
 __blockdev_direct_IO(int rw, struct kiocb *iocb, struct inode *inode,
@@ -1211,6 +1105,10 @@ __blockdev_direct_IO(int rw, struct kiocb *iocb, struct inode *inode,
        ssize_t retval = -EINVAL;
        loff_t end = offset;
        struct dio *dio;
+       struct dio_submit sdio = { 0, };
+       unsigned long user_addr;
+       size_t bytes;
+       struct buffer_head map_bh = { 0, };
 
        if (rw & WRITE)
                rw = WRITE_ODIRECT;
@@ -1244,7 +1142,7 @@ __blockdev_direct_IO(int rw, struct kiocb *iocb, struct inode *inode,
        if (rw == READ && end == offset)
                return 0;
 
-       dio = kmalloc(sizeof(*dio), GFP_KERNEL);
+       dio = kmem_cache_alloc(dio_cache, GFP_KERNEL);
        retval = -ENOMEM;
        if (!dio)
                goto out;
@@ -1268,7 +1166,7 @@ __blockdev_direct_IO(int rw, struct kiocb *iocb, struct inode *inode,
                                                              end - 1);
                        if (retval) {
                                mutex_unlock(&inode->i_mutex);
-                               kfree(dio);
+                               kmem_cache_free(dio_cache, dio);
                                goto out;
                        }
                }
@@ -1288,11 +1186,141 @@ __blockdev_direct_IO(int rw, struct kiocb *iocb, struct inode *inode,
        dio->is_async = !is_sync_kiocb(iocb) && !((rw & WRITE) &&
                (end > i_size_read(inode)));
 
-       retval = direct_io_worker(rw, iocb, inode, iov, offset,
-                               nr_segs, blkbits, get_block, end_io,
-                               submit_io, dio);
+       retval = 0;
+
+       dio->inode = inode;
+       dio->rw = rw;
+       sdio.blkbits = blkbits;
+       sdio.blkfactor = inode->i_blkbits - blkbits;
+       sdio.block_in_file = offset >> blkbits;
+
+       sdio.get_block = get_block;
+       dio->end_io = end_io;
+       sdio.submit_io = submit_io;
+       sdio.final_block_in_bio = -1;
+       sdio.next_block_for_io = -1;
+
+       dio->iocb = iocb;
+       dio->i_size = i_size_read(inode);
+
+       spin_lock_init(&dio->bio_lock);
+       dio->refcount = 1;
+
+       /*
+        * In case of non-aligned buffers, we may need 2 more
+        * pages since we need to zero out first and last block.
+        */
+       if (unlikely(sdio.blkfactor))
+               sdio.pages_in_io = 2;
+
+       for (seg = 0; seg < nr_segs; seg++) {
+               user_addr = (unsigned long)iov[seg].iov_base;
+               sdio.pages_in_io +=
+                       ((user_addr + iov[seg].iov_len + PAGE_SIZE-1) /
+                               PAGE_SIZE - user_addr / PAGE_SIZE);
+       }
+
+       for (seg = 0; seg < nr_segs; seg++) {
+               user_addr = (unsigned long)iov[seg].iov_base;
+               sdio.size += bytes = iov[seg].iov_len;
+
+               /* Index into the first page of the first block */
+               sdio.first_block_in_page = (user_addr & ~PAGE_MASK) >> blkbits;
+               sdio.final_block_in_request = sdio.block_in_file +
+                                               (bytes >> blkbits);
+               /* Page fetching state */
+               sdio.head = 0;
+               sdio.tail = 0;
+               sdio.curr_page = 0;
+
+               sdio.total_pages = 0;
+               if (user_addr & (PAGE_SIZE-1)) {
+                       sdio.total_pages++;
+                       bytes -= PAGE_SIZE - (user_addr & (PAGE_SIZE - 1));
+               }
+               sdio.total_pages += (bytes + PAGE_SIZE - 1) / PAGE_SIZE;
+               sdio.curr_user_address = user_addr;
+
+               retval = do_direct_IO(dio, &sdio, &map_bh);
+
+               dio->result += iov[seg].iov_len -
+                       ((sdio.final_block_in_request - sdio.block_in_file) <<
+                                       blkbits);
+
+               if (retval) {
+                       dio_cleanup(dio, &sdio);
+                       break;
+               }
+       } /* end iovec loop */
+
+       if (retval == -ENOTBLK) {
+               /*
+                * The remaining part of the request will be
+                * be handled by buffered I/O when we return
+                */
+               retval = 0;
+       }
+       /*
+        * There may be some unwritten disk at the end of a part-written
+        * fs-block-sized block.  Go zero that now.
+        */
+       dio_zero_block(dio, &sdio, 1, &map_bh);
+
+       if (sdio.cur_page) {
+               ssize_t ret2;
+
+               ret2 = dio_send_cur_page(dio, &sdio, &map_bh);
+               if (retval == 0)
+                       retval = ret2;
+               page_cache_release(sdio.cur_page);
+               sdio.cur_page = NULL;
+       }
+       if (sdio.bio)
+               dio_bio_submit(dio, &sdio);
+
+       /*
+        * It is possible that, we return short IO due to end of file.
+        * In that case, we need to release all the pages we got hold on.
+        */
+       dio_cleanup(dio, &sdio);
+
+       /*
+        * All block lookups have been performed. For READ requests
+        * we can let i_mutex go now that its achieved its purpose
+        * of protecting us from looking up uninitialized blocks.
+        */
+       if (rw == READ && (dio->flags & DIO_LOCKING))
+               mutex_unlock(&dio->inode->i_mutex);
+
+       /*
+        * The only time we want to leave bios in flight is when a successful
+        * partial aio read or full aio write have been setup.  In that case
+        * bio completion will call aio_complete.  The only time it's safe to
+        * call aio_complete is when we return -EIOCBQUEUED, so we key on that.
+        * This had *better* be the only place that raises -EIOCBQUEUED.
+        */
+       BUG_ON(retval == -EIOCBQUEUED);
+       if (dio->is_async && retval == 0 && dio->result &&
+           ((rw & READ) || (dio->result == sdio.size)))
+               retval = -EIOCBQUEUED;
+
+       if (retval != -EIOCBQUEUED)
+               dio_await_completion(dio);
+
+       if (drop_refcount(dio) == 0) {
+               retval = dio_complete(dio, offset, retval, false);
+               kmem_cache_free(dio_cache, dio);
+       } else
+               BUG_ON(retval != -EIOCBQUEUED);
 
 out:
        return retval;
 }
 EXPORT_SYMBOL(__blockdev_direct_IO);
+
+static __init int dio_init(void)
+{
+       dio_cache = KMEM_CACHE(dio, SLAB_PANIC);
+       return 0;
+}
+module_init(dio_init)
index e4095e9..b9548f4 100644 (file)
@@ -224,53 +224,8 @@ loff_t ext4_llseek(struct file *file, loff_t offset, int origin)
                maxbytes = EXT4_SB(inode->i_sb)->s_bitmap_maxbytes;
        else
                maxbytes = inode->i_sb->s_maxbytes;
-       mutex_lock(&inode->i_mutex);
-       switch (origin) {
-       case SEEK_END:
-               offset += inode->i_size;
-               break;
-       case SEEK_CUR:
-               if (offset == 0) {
-                       mutex_unlock(&inode->i_mutex);
-                       return file->f_pos;
-               }
-               offset += file->f_pos;
-               break;
-       case SEEK_DATA:
-               /*
-                * In the generic case the entire file is data, so as long as
-                * offset isn't at the end of the file then the offset is data.
-                */
-               if (offset >= inode->i_size) {
-                       mutex_unlock(&inode->i_mutex);
-                       return -ENXIO;
-               }
-               break;
-       case SEEK_HOLE:
-               /*
-                * There is a virtual hole at the end of the file, so as long as
-                * offset isn't i_size or larger, return i_size.
-                */
-               if (offset >= inode->i_size) {
-                       mutex_unlock(&inode->i_mutex);
-                       return -ENXIO;
-               }
-               offset = inode->i_size;
-               break;
-       }
-
-       if (offset < 0 || offset > maxbytes) {
-               mutex_unlock(&inode->i_mutex);
-               return -EINVAL;
-       }
-
-       if (offset != file->f_pos) {
-               file->f_pos = offset;
-               file->f_version = 0;
-       }
-       mutex_unlock(&inode->i_mutex);
 
-       return offset;
+       return generic_file_llseek_size(file, offset, origin, maxbytes);
 }
 
 const struct file_operations ext4_file_operations = {
index 5002408..ce36a56 100644 (file)
@@ -66,13 +66,13 @@ static loff_t gfs2_llseek(struct file *file, loff_t offset, int origin)
                error = gfs2_glock_nq_init(ip->i_gl, LM_ST_SHARED, LM_FLAG_ANY,
                                           &i_gh);
                if (!error) {
-                       error = generic_file_llseek_unlocked(file, offset, origin);
+                       error = generic_file_llseek(file, offset, origin);
                        gfs2_glock_dq_uninit(&i_gh);
                }
                break;
        case SEEK_CUR:
        case SEEK_SET:
-               error = generic_file_llseek_unlocked(file, offset, origin);
+               error = generic_file_llseek(file, offset, origin);
                break;
        default:
                error = -EINVAL;
index ec79246..ecbb68d 100644 (file)
@@ -634,7 +634,7 @@ void prune_icache_sb(struct super_block *sb, int nr_to_scan)
                 * inode to the back of the list so we don't spin on it.
                 */
                if (!spin_trylock(&inode->i_lock)) {
-                       list_move(&inode->i_lru, &sb->s_inode_lru);
+                       list_move_tail(&inode->i_lru, &sb->s_inode_lru);
                        continue;
                }
 
index 0b3138d..7657be4 100644 (file)
@@ -221,14 +221,12 @@ static int check_acl(struct inode *inode, int mask)
 }
 
 /*
- * This does basic POSIX ACL permission checking
+ * This does the basic permission checking
  */
 static int acl_permission_check(struct inode *inode, int mask)
 {
        unsigned int mode = inode->i_mode;
 
-       mask &= MAY_READ | MAY_WRITE | MAY_EXEC | MAY_NOT_BLOCK;
-
        if (current_user_ns() != inode_userns(inode))
                goto other_perms;
 
@@ -257,7 +255,7 @@ other_perms:
 /**
  * generic_permission -  check for access rights on a Posix-like filesystem
  * @inode:     inode to check access rights for
- * @mask:      right to check for (%MAY_READ, %MAY_WRITE, %MAY_EXEC)
+ * @mask:      right to check for (%MAY_READ, %MAY_WRITE, %MAY_EXEC, ...)
  *
  * Used to check for read/write/execute permissions on a file.
  * We use "fsuid" for this, letting us set arbitrary permissions
@@ -273,7 +271,7 @@ int generic_permission(struct inode *inode, int mask)
        int ret;
 
        /*
-        * Do the basic POSIX ACL permission checks.
+        * Do the basic permission checks.
         */
        ret = acl_permission_check(inode, mask);
        if (ret != -EACCES)
@@ -331,12 +329,14 @@ static inline int do_inode_permission(struct inode *inode, int mask)
 /**
  * inode_permission  -  check for access rights to a given inode
  * @inode:     inode to check permission on
- * @mask:      right to check for (%MAY_READ, %MAY_WRITE, %MAY_EXEC)
+ * @mask:      right to check for (%MAY_READ, %MAY_WRITE, %MAY_EXEC, ...)
  *
  * Used to check for read/write/execute permissions on an inode.
  * We use "fsuid" for this, letting us set arbitrary permissions
  * for filesystem access without changing the "normal" uids which
  * are used for other things.
+ *
+ * When checking for MAY_APPEND, MAY_WRITE must also be set in @mask.
  */
 int inode_permission(struct inode *inode, int mask)
 {
@@ -2035,10 +2035,7 @@ static int may_open(struct path *path, int acc_mode, int flag)
        if (flag & O_NOATIME && !inode_owner_or_capable(inode))
                return -EPERM;
 
-       /*
-        * Ensure there are no outstanding leases on the file.
-        */
-       return break_lease(inode, flag);
+       return 0;
 }
 
 static int handle_truncate(struct file *filp)
index b4febb2..e5e1c7d 100644 (file)
@@ -1109,6 +1109,7 @@ static int show_vfsstat(struct seq_file *m, void *v)
 
        /* device */
        if (mnt->mnt_sb->s_op->show_devname) {
+               seq_puts(m, "device ");
                err = mnt->mnt_sb->s_op->show_devname(m, mnt);
        } else {
                if (mnt->mnt_devname) {
index 28b8c3f..91c01f0 100644 (file)
@@ -180,8 +180,6 @@ force_reval:
 
 static loff_t nfs_file_llseek(struct file *filp, loff_t offset, int origin)
 {
-       loff_t loff;
-
        dprintk("NFS: llseek file(%s/%s, %lld, %d)\n",
                        filp->f_path.dentry->d_parent->d_name.name,
                        filp->f_path.dentry->d_name.name,
@@ -197,13 +195,9 @@ static loff_t nfs_file_llseek(struct file *filp, loff_t offset, int origin)
                int retval = nfs_revalidate_file_size(inode, filp);
                if (retval < 0)
                        return (loff_t)retval;
+       }
 
-               spin_lock(&inode->i_lock);
-               loff = generic_file_llseek_unlocked(filp, offset, origin);
-               spin_unlock(&inode->i_lock);
-       } else
-               loff = generic_file_llseek_unlocked(filp, offset, origin);
-       return loff;
+       return generic_file_llseek(filp, offset, origin);
 }
 
 /*
index f711921..22c41b5 100644 (file)
--- a/fs/open.c
+++ b/fs/open.c
@@ -685,6 +685,10 @@ static struct file *__dentry_open(struct dentry *dentry, struct vfsmount *mnt,
        if (error)
                goto cleanup_all;
 
+       error = break_lease(inode, f->f_flags);
+       if (error)
+               goto cleanup_all;
+
        if (!open && f->f_op)
                open = f->f_op->open;
        if (open) {
index 10027b4..cea4623 100644 (file)
@@ -218,6 +218,8 @@ posix_acl_permission(struct inode *inode, const struct posix_acl *acl, int want)
        const struct posix_acl_entry *pa, *pe, *mask_obj;
        int found = 0;
 
+       want &= MAY_READ | MAY_WRITE | MAY_EXEC | MAY_NOT_BLOCK;
+
        FOREACH_ACL_ENTRY(pa, acl, pe) {
                 switch(pa->e_tag) {
                         case ACL_USER_OBJ:
index 179f1c3..dfd1257 100644 (file)
@@ -35,23 +35,45 @@ static inline int unsigned_offsets(struct file *file)
        return file->f_mode & FMODE_UNSIGNED_OFFSET;
 }
 
+static loff_t lseek_execute(struct file *file, struct inode *inode,
+               loff_t offset, loff_t maxsize)
+{
+       if (offset < 0 && !unsigned_offsets(file))
+               return -EINVAL;
+       if (offset > maxsize)
+               return -EINVAL;
+
+       if (offset != file->f_pos) {
+               file->f_pos = offset;
+               file->f_version = 0;
+       }
+       return offset;
+}
+
 /**
* generic_file_llseek_unlocked - lockless generic llseek implementation
  * @file:      file structure to seek on
  * @offset:    file offset to seek to
  * @origin:    type of seek
+ * @size:      max size of file system
+ *
+ * This is a variant of generic_file_llseek that allows passing in a custom
+ * file size.
  *
- * Updates the file offset to the value specified by @offset and @origin.
- * Locking must be provided by the caller.
+ * Synchronization:
+ * SEEK_SET and SEEK_END are unsynchronized (but atomic on 64bit platforms)
+ * SEEK_CUR is synchronized against other SEEK_CURs, but not read/writes.
+ * read/writes behave like SEEK_SET against seeks.
  */
 loff_t
-generic_file_llseek_unlocked(struct file *file, loff_t offset, int origin)
+generic_file_llseek_size(struct file *file, loff_t offset, int origin,
+               loff_t maxsize)
 {
        struct inode *inode = file->f_mapping->host;
 
        switch (origin) {
        case SEEK_END:
-               offset += inode->i_size;
+               offset += i_size_read(inode);
                break;
        case SEEK_CUR:
                /*
@@ -62,14 +84,22 @@ generic_file_llseek_unlocked(struct file *file, loff_t offset, int origin)
                 */
                if (offset == 0)
                        return file->f_pos;
-               offset += file->f_pos;
-               break;
+               /*
+                * f_lock protects against read/modify/write race with other
+                * SEEK_CURs. Note that parallel writes and reads behave
+                * like SEEK_SET.
+                */
+               spin_lock(&file->f_lock);
+               offset = lseek_execute(file, inode, file->f_pos + offset,
+                                      maxsize);
+               spin_unlock(&file->f_lock);
+               return offset;
        case SEEK_DATA:
                /*
                 * In the generic case the entire file is data, so as long as
                 * offset isn't at the end of the file then the offset is data.
                 */
-               if (offset >= inode->i_size)
+               if (offset >= i_size_read(inode))
                        return -ENXIO;
                break;
        case SEEK_HOLE:
@@ -77,26 +107,15 @@ generic_file_llseek_unlocked(struct file *file, loff_t offset, int origin)
                 * There is a virtual hole at the end of the file, so as long as
                 * offset isn't i_size or larger, return i_size.
                 */
-               if (offset >= inode->i_size)
+               if (offset >= i_size_read(inode))
                        return -ENXIO;
-               offset = inode->i_size;
+               offset = i_size_read(inode);
                break;
        }
 
-       if (offset < 0 && !unsigned_offsets(file))
-               return -EINVAL;
-       if (offset > inode->i_sb->s_maxbytes)
-               return -EINVAL;
-
-       /* Special lock needed here? */
-       if (offset != file->f_pos) {
-               file->f_pos = offset;
-               file->f_version = 0;
-       }
-
-       return offset;
+       return lseek_execute(file, inode, offset, maxsize);
 }
-EXPORT_SYMBOL(generic_file_llseek_unlocked);
+EXPORT_SYMBOL(generic_file_llseek_size);
 
 /**
  * generic_file_llseek - generic llseek implementation for regular files
@@ -110,13 +129,10 @@ EXPORT_SYMBOL(generic_file_llseek_unlocked);
  */
 loff_t generic_file_llseek(struct file *file, loff_t offset, int origin)
 {
-       loff_t rval;
-
-       mutex_lock(&file->f_dentry->d_inode->i_mutex);
-       rval = generic_file_llseek_unlocked(file, offset, origin);
-       mutex_unlock(&file->f_dentry->d_inode->i_mutex);
+       struct inode *inode = file->f_mapping->host;
 
-       return rval;
+       return generic_file_llseek_size(file, offset, origin,
+                                       inode->i_sb->s_maxbytes);
 }
 EXPORT_SYMBOL(generic_file_llseek);
 
index ba98668..14493a2 100644 (file)
@@ -58,14 +58,15 @@ struct inodes_stat_t {
 
 #define NR_FILE  8192  /* this can well be larger on a larger system */
 
-#define MAY_EXEC 1
-#define MAY_WRITE 2
-#define MAY_READ 4
-#define MAY_APPEND 8
-#define MAY_ACCESS 16
-#define MAY_OPEN 32
-#define MAY_CHDIR 64
-#define MAY_NOT_BLOCK 128      /* called from RCU mode, don't block */
+#define MAY_EXEC               0x00000001
+#define MAY_WRITE              0x00000002
+#define MAY_READ               0x00000004
+#define MAY_APPEND             0x00000008
+#define MAY_ACCESS             0x00000010
+#define MAY_OPEN               0x00000020
+#define MAY_CHDIR              0x00000040
+/* called from RCU mode, don't block */
+#define MAY_NOT_BLOCK          0x00000080
 
 /*
  * flags in file.f_mode.  Note that FMODE_READ and FMODE_WRITE must correspond
@@ -963,7 +964,12 @@ struct file {
 #define f_dentry       f_path.dentry
 #define f_vfsmnt       f_path.mnt
        const struct file_operations    *f_op;
-       spinlock_t              f_lock;  /* f_ep_links, f_flags, no IRQ */
+
+       /*
+        * Protects f_ep_links, f_flags, f_pos vs i_size in lseek SEEK_CUR.
+        * Must not be taken from IRQ context.
+        */
+       spinlock_t              f_lock;
 #ifdef CONFIG_SMP
        int                     f_sb_list_cpu;
 #endif
@@ -2401,8 +2407,8 @@ file_ra_state_init(struct file_ra_state *ra, struct address_space *mapping);
 extern loff_t noop_llseek(struct file *file, loff_t offset, int origin);
 extern loff_t no_llseek(struct file *file, loff_t offset, int origin);
 extern loff_t generic_file_llseek(struct file *file, loff_t offset, int origin);
-extern loff_t generic_file_llseek_unlocked(struct file *file, loff_t offset,
-                       int origin);
+extern loff_t generic_file_llseek_size(struct file *file, loff_t offset,
+               int origin, loff_t maxsize);
 extern int generic_file_open(struct inode * inode, struct file * filp);
 extern int nonseekable_open(struct inode * inode, struct file * filp);
 
index 7771871..5cf820a 100644 (file)
@@ -2115,6 +2115,7 @@ void iov_iter_advance(struct iov_iter *i, size_t bytes)
        } else {
                const struct iovec *iov = i->iov;
                size_t base = i->iov_offset;
+               unsigned long nr_segs = i->nr_segs;
 
                /*
                 * The !iov->iov_len check ensures we skip over unlikely
@@ -2130,11 +2131,13 @@ void iov_iter_advance(struct iov_iter *i, size_t bytes)
                        base += copy;
                        if (iov->iov_len == base) {
                                iov++;
+                               nr_segs--;
                                base = 0;
                        }
                }
                i->iov = iov;
                i->iov_offset = base;
+               i->nr_segs = nr_segs;
        }
 }
 EXPORT_SYMBOL(iov_iter_advance);
@@ -35,23 +35,45 @@ static inline int unsigned_offsets(struct file *file)
 /**
- * generic_file_llseek_unlocked - lockless generic llseek implementation
+ * generic_file_llseek_size - generic llseek implementation for regular files
  * @file:      file structure to seek on
  * @offset:    file offset to seek to